/* Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <my_global.h>
#include <my_sys.h>
#include <assert.h>
#include <errno.h>
#include "m_string.h"


/* VMT for cache files.. */
MY_FILE_VMT cache_file_vmt= 
{
  &simple_file_vmt,
  &my_cache_read,
  &my_simple_pread,
  &my_cache_write,
  &my_simple_pwrite,
  &my_cache_close,
  &my_simple_lock,
  &my_cache_chsize,
  &my_simple_fstat,
  &my_cache_tell,
  &my_cache_seek,
  &my_cache_free
};


/* VMT for cache files.. */
MY_FILE_VMT cache_read_file_vmt= 
{
  &simple_file_vmt,
  &my_cache_read,
  &my_simple_pread,
  &my_simple_write,
  &my_simple_pwrite,
  &my_cache_close,
  &my_simple_lock,
  &my_cache_chsize,
  &my_simple_fstat,
  &my_cache_tell,
  &my_cache_seek,
  &my_cache_free
};


/* VMT for cache files.. */
MY_FILE_VMT cache_write_file_vmt= 
{
  &simple_file_vmt,
  &my_simple_read,
  &my_simple_pread,
  &my_cache_write,
  &my_simple_pwrite,
  &my_cache_close,
  &my_simple_lock,
  &my_cache_chsize,
  &my_simple_fstat,
  &my_cache_tell,
  &my_cache_seek,
  &my_cache_free
};


/* VMT for cache files with SEQ_READ_APPEND.. */
MY_FILE_VMT cache_seq_read_append_file_vmt= 
{
  &simple_file_vmt,
  &my_cache_seq_read,
  &my_simple_pread,
  0,			/* Force a core if used */
  0,			/* Force a core if used */
  &my_cache_close,
  &my_simple_lock,
  &my_cache_chsize,
  &my_simple_fstat,
  &my_cache_tell,
  &my_cache_seek,
  &my_cache_free
};


MY_FILE_VMT cache_raid_file_vmt= 
{
  &raid_file_vmt,
  &my_cache_read,
  &my_u_raid_pread,
  &my_cache_write,
  &my_u_raid_pwrite,
  &my_cache_close,
  &my_u_raid_lock,
  &my_cache_chsize,
  &my_u_raid_fstat,
  &my_cache_tell,
  &my_cache_seek,
  &my_cache_free
};


/* VMT for cache files.. */
MY_FILE_VMT cache_read_raid_file_vmt= 
{
  &raid_file_vmt,
  &my_cache_read,
  &my_u_raid_pread,
  &my_u_raid_write,
  &my_u_raid_pwrite,
  &my_cache_close,
  &my_u_raid_lock,
  &my_cache_chsize,
  &my_u_raid_fstat,
  &my_cache_tell,
  &my_cache_seek,
  &my_cache_free
};


/* VMT for cache files.. */
MY_FILE_VMT cache_write_raid_file_vmt= 
{
  &raid_file_vmt,
  &my_u_raid_read,
  &my_u_raid_pread,
  &my_cache_write,
  &my_u_raid_pwrite,
  &my_cache_close,
  &my_u_raid_lock,
  &my_cache_chsize,
  &my_u_raid_fstat,
  &my_cache_tell,
  &my_cache_seek,
  &my_cache_free
};


/* VMT for cache files with SEQ_READ_APPEND.. */
MY_FILE_VMT cache_seq_read_append_raid_file_vmt= 
{
  &raid_file_vmt,
  &my_cache_seq_read,
  &my_u_raid_pread,
  0,			/* Force a core if used */
  0,			/* Force a core if used */
  &my_cache_close,
  &my_u_raid_lock,
  &my_cache_chsize,
  &my_u_raid_fstat,
  &my_cache_tell,
  &my_cache_seek,
  &my_cache_free
};


#ifdef HAVE_AIOWAIT
/* VMT for cache files with asynchronous read.. */
MY_FILE_VMT cache_file_vmt_async_read= 
{
  &simple_file_vmt,
  &my_cache_read_async,
  &my_simple_pread,
  &my_cache_write,
  &my_simple_pwrite,
  &my_cache_close,
  &my_simple_lock,
  &my_cache_chsize,
  &my_simple_fstat,
  &my_cache_tell,
  &my_cache_seek,
  &my_cache_free
};


/* VMT for cache files with asynchronous read.. */
MY_FILE_VMT cache_raid_file_vmt_async_read= 
{
  &raid_file_vmt,
  &my_cache_read_async,
  &my_u_raid_pread,
  &my_cache_write,
  &my_u_raid_pwrite,
  &my_cache_close,
  &my_u_raid_lock,
  &my_cache_chsize,
  &my_u_raid_fstat,
  &my_cache_tell,
  &my_cache_seek,
  &my_cache_free
};

#endif

void my_cache_free(MY_FILE *my_file)
{
}

static void init_u_functions(MY_FILE* file, enum en_cache_type type)
{
  MY_FILE_CACHE * info = &file->cache;
  DBUG_ENTER("init_u_functions");

  if (file->raid_root.raid_type != STRIPPED)
  {
    switch (type) {
    case EN_READ_NET:
      /* 
	 must be initialized by the caller. The problem is that
	 _my_b_net_read has to be defined in sql directory because of
	 the dependency on THD, and therefore cannot be visible to
	 programs that link against mysys but know nothing about THD, such
	 as myisamchk
      */
      break;
    case EN_SEQ_READ_APPEND:
      file->vmt = &cache_seq_read_append_file_vmt;
      break;
    case EN_WRITE_CACHE:
      file->vmt = &cache_write_file_vmt;
      break;
    case EN_READ_CACHE:
      file->vmt = &cache_read_file_vmt;
      break;
    default:
      file->vmt = &cache_file_vmt;
    }
  }
  else
  {
    switch (type) {
    case EN_READ_NET:
      /* 
	 must be initialized by the caller. The problem is that
	 _my_b_net_read has to be defined in sql directory because of
	 the dependency on THD, and therefore cannot be visible to
	 programs that link against mysys but know nothing about THD, such
	 as myisamchk
      */
      break;
    case EN_SEQ_READ_APPEND:
      file->vmt = &cache_seq_read_append_raid_file_vmt;
      break;
    case EN_WRITE_CACHE:
      file->vmt = &cache_write_raid_file_vmt;
      break;
    case EN_READ_CACHE:
      file->vmt = &cache_read_raid_file_vmt;
      break;
    default:
      file->vmt = &cache_raid_file_vmt;
    }
  }

  /* Ensure that my_b_tell() and my_b_bytes_in_cache works */
  if (type == EN_WRITE_CACHE)
  {
    info->current_pos= &info->write_pos;
    info->current_end= &info->write_end;
  }
  else
  {
    info->current_pos= &info->read_pos;
    info->current_end= &info->read_end;
  }
  DBUG_VOID_RETURN;
}


/*
  if cachesize == 0 then use default cachesize (from s-file)
  if file == -1 then real_open_cached_file() will be called.
  returns 0 if ok
*/
extern int init_u_cache(MY_FILE *file, uint cachesize,
			enum en_cache_type type, my_off_t seek_offset,
			pbool use_async_io, myf cache_myflags)
{
  MY_FILE_CACHE *info= &file->cache;
  uint min_cache;
  my_off_t end_of_file= ~(my_off_t) 0;
  DBUG_ENTER("init_u_cache");
  DBUG_PRINT("enter",("ptr=%p fd=%d type: %d  pos: %ld",
		      file,(int)file->file_descr,
		      (int) type, (ulong) seek_offset));

  info->type=type;
  file->pos_in_file= seek_offset;
  info->pre_close = info->pre_read = info->post_read = 0;
  info->arg = 0;
  info->alloced_buffer = 0;
  info->buffer=0;
  info->seek_not_done= test(file->file_descr >= 0);

  if (!cachesize)
    if (! (cachesize= my_default_record_cache_size))
      DBUG_RETURN(1);				/* No cache requested */
  min_cache=use_async_io ? IO_SIZE*4 : IO_SIZE*2;
  if (type == EN_READ_CACHE || type == EN_SEQ_READ_APPEND)
  {						/* Assume file isn't growing */
    if (!(cache_myflags & MY_DONT_CHECK_FILESIZE))
    {
      /* Calculate end of file to not allocate to big buffers */
      end_of_file=my_u_seek(file,0L,MY_SEEK_END,MYF(0));
      if (end_of_file < seek_offset)
	end_of_file=seek_offset;
      /* Trim cache size if the file is very small */
      if ((my_off_t) cachesize > end_of_file-seek_offset+IO_SIZE*2-1)
      {
	cachesize=(uint) (end_of_file-seek_offset)+IO_SIZE*2-1;
	use_async_io=0;				/* No need to use async */
      }
    }
  }
  cache_myflags &= ~MY_DONT_CHECK_FILESIZE;
  if (type != EN_READ_NET && type != EN_WRITE_NET)
  {
    /* Retry allocating memory in smaller blocks until we get one */
    for (;;)
    {
      uint buffer_block;
      cachesize=(uint) ((ulong) (cachesize + min_cache-1) &
			(ulong) ~(min_cache-1));
      if (cachesize < min_cache)
	cachesize = min_cache;
      buffer_block = cachesize;
      if (type == EN_SEQ_READ_APPEND)
	buffer_block *= 2;
      if ((info->buffer=
	   (byte*) my_malloc(buffer_block,
			     MYF((cache_myflags & ~ MY_WME) |
				 (cachesize == min_cache ? MY_WME : 0)))) != 0)
      {
	info->write_buffer=info->buffer;
	if (type == EN_SEQ_READ_APPEND)
	  info->write_buffer = info->buffer + cachesize;
	info->alloced_buffer=1;
	break;					/* Enough memory found */
      }
      if (cachesize == min_cache)
	DBUG_RETURN(2);				/* Can't alloc cache */
      cachesize= (uint) ((long) cachesize*3/4); /* Try with less memory */
    }
  }
  
  DBUG_PRINT("info",("init_u_cache: cachesize = %u",cachesize));
  info->read_length=info->buffer_length=cachesize;
  file->myflags=cache_myflags & ~(MY_NABP | MY_FNABP);
  info->request_pos= info->read_pos= info->write_pos = info->buffer;
  if (type == EN_SEQ_READ_APPEND)
  {
    info->append_read_pos = info->write_pos = info->write_buffer;
    info->write_end = info->write_buffer + info->buffer_length;
#ifdef THREAD
    pthread_mutex_init(&info->append_buffer_lock,MY_MUTEX_INIT_FAST);
#endif
  }
  
  if (type == EN_WRITE_CACHE)
    info->write_end=
      info->buffer+info->buffer_length- (seek_offset & (IO_SIZE-1));
  else
    info->read_end=info->buffer;		/* Nothing in cache */
  
  /* End_of_file may be changed by user later */
  file->end_of_file= end_of_file;
  info->error=0;
  init_u_functions(file,type);
#ifdef HAVE_AIOWAIT
  if (use_async_io && ! my_disable_async_io)
  {
    DBUG_PRINT("info",("Using async io"));
    info->read_length/=2;
    info->read_function=_my_b_async_read;
  }
  info->inited=info->aio_result.pending=0;
#endif
  DBUG_RETURN(0);
}


my_size_t my_cache_read(MY_FILE *file, byte *Buffer, my_size_t Count, 
			myf myflags __attribute__((unused)))
{
  int begin_count = Count;
  MY_FILE_CACHE *info= &file->cache;
  uint length,diff_length,left_length;
  my_off_t max_length, pos_in_file;
  DBUG_ENTER("my_cache_read");
  
  if ((left_length=(uint) (info->read_end-info->read_pos)))
  {
    DBUG_ASSERT(Count >= left_length);	/* User is not using my_b_read() */
    memcpy(Buffer,info->read_pos, (my_size_t) (left_length));
    Buffer+=left_length;
    Count-=left_length;
  }
  
  /* pos_in_file always point on where info->buffer was read */
  pos_in_file=file->pos_in_file+(uint) (info->read_end - info->buffer);
  if (info->seek_not_done)
  {					/* File touched, do seek */
    VOID(my_u_super_seek(file, pos_in_file, MY_SEEK_SET,MYF(0)));
    info->seek_not_done=0;
  }
  diff_length=(uint) (pos_in_file & (IO_SIZE-1));
  if (Count >= (uint) (IO_SIZE+(IO_SIZE-diff_length)))
  {					/* Fill first intern buffer */
    uint read_length;
    if (file->end_of_file == pos_in_file)
    {					/* End of file */
      info->error=(int) left_length;
      DBUG_RETURN((my_size_t)-1);
    }
    length=(Count & (uint) ~(IO_SIZE-1))-diff_length;
    if ((read_length=my_u_super_read(file,Buffer,length,file->myflags))
	!= length)
    {
      info->error= read_length == (uint) -1 ? -1 :
	(int) (read_length+left_length);
      DBUG_RETURN((my_size_t)-1);
    }
    Count-=length;
    Buffer+=length;
    pos_in_file+=length;
    left_length+=length;
    diff_length=0;
  }
  
  max_length=info->read_length-diff_length;
  if (info->type != EN_READ_FIFO &&
      max_length > (file->end_of_file - pos_in_file))
    max_length = file->end_of_file - pos_in_file;
  if (!max_length)
  {
    if (Count)
    {
      info->error= left_length;		/* We only got this many char */
      DBUG_RETURN((my_size_t)-1);
    }
    length=0;				/* Didn't read any chars */
  }
  else if ((length=my_u_super_read(file,info->buffer,(uint) max_length,
				   file->myflags)) < Count ||
	   length == (uint) -1)
  {
    if (length != (uint) -1)
      memcpy(Buffer,info->buffer,(my_size_t) length);
    info->error= length == (uint) -1 ? -1 : (int) (length+left_length);
    info->read_pos=info->read_end=info->buffer;
    DBUG_RETURN((my_size_t)-1);
  }
  info->read_pos=info->buffer+Count;
  info->read_end=info->buffer+length;
  file->pos_in_file=pos_in_file;
  memcpy(Buffer,info->buffer,(my_size_t) Count);
  DBUG_RETURN(begin_count);
}


/*
  Write a block to disk where part of the data may be inside the record
  buffer.  As all write calls to the data goes through the cache,
  we will never get a seek over the end of the buffer
*/
int my_u_block_write(register MY_FILE *file, const byte *Buffer, 
		     my_size_t Count, my_off_t pos, myf myflags)
{
  MY_FILE_CACHE *info= &file->cache;
  uint length;
  int error=0;
  DBUG_ENTER("my_u_block_write");
  
  if (pos < file->pos_in_file)
  {
    /* Of no overlap, write everything without buffering */
    if (pos + Count <= file->pos_in_file)
      DBUG_RETURN(my_u_super_pwrite(file, Buffer, Count, pos,
				    file->myflags | MY_NABP)!=Count);
    /* Write the part of the block that is before buffer */
    length= (uint) (file->pos_in_file - pos);
    if (my_u_super_pwrite(file, Buffer, length, pos, 
			  file->myflags|MY_NABP)!=length)
      info->error=error=-1;
    Buffer+=length;
    pos+=  length;
    Count-= length;
#ifndef HAVE_PREAD
    info->seek_not_done=1;
#endif
  }
  
  /* Check if we want to write inside the used part of the buffer.*/
  length= (uint) (info->write_end - info->buffer);
  if (pos < file->pos_in_file + length)
  {
    uint offset= (uint) (pos - file->pos_in_file);
    length-=offset;
    if (length > Count)
      length=Count;
    memcpy(info->buffer+offset, Buffer, length);
    Buffer+=length;
    Count-= length;
    /* Fix length of buffer if the new data was larger */
    if (info->buffer+length > info->write_pos)
      info->write_pos=info->buffer+length;
    if (!Count)
      DBUG_RETURN(error);
  }
  /* Write at the end of the current buffer; This is the normal case */
  if (my_cache_write(file, Buffer, Count, myflags)!=Count)
    error= -1;
  DBUG_RETURN(error);
}


#ifdef HAVE_AIOWAIT
int my_cache_read_async	(MY_FILE *fs, byte *Buffer, uint Count, myf myflags)
{  
  uint length,read_length,diff_length,left_length,use_length,org_Count;
  my_off_t max_length;
  my_off_t next_pos_in_file;
  byte *read_buffer;
  DBUG_ENTER("my_cache_read_async");
  
  memcpy(Buffer,fs->rc_pos,
	 (my_size_t) (left_length= (uint) (fs->rc_end - fs->rc_pos)));
  Buffer+= left_length;
  org_Count= Count;
  Count-= left_length;
  
  if (fs->inited)
  {						/* wait for read block */
    fs->inited= 0;				/* No more block to read */
    my_aiowait(&fs->aio_result);		/* Wait for outstanding req */
    if (fs->aio_result.result.aio_errno)
    {
      if (fs->myflags & MY_WME)
	my_error(EE_READ, MYF(ME_BELL+ME_WAITTANG),
		 my_u_filename(fs->file),
		 fs->aio_result.result.aio_errno);
      my_errno= fs->aio_result.result.aio_errno;
      fs->error= -1;
      DBUG_RETURN(1);
    }
    if (! (read_length= (uint) fs->aio_result.result.aio_return) ||
	read_length == (uint) -1)
    {
      my_errno= 0;				/* For testing */
      fs->error= (read_length == (uint) -1 ? -1 :
		  (int) (read_length+left_length));
      DBUG_RETURB(1);
    }
    fs->pos_in_file+= (uint) (fs->rc_end - fs->rc_request_pos);
    
    if (fs->rc_request_pos != fs->buffer)
      fs->rc_request_pos= fs->buffer;
    else
      fs->rc_request_pos= fs->buffer + fs->read_length;
    fs->rc_pos= fs->rc_request_pos;
    next_pos_in_file= fs->aio_read_pos + read_length;
    
    /* Check if pos_in_file is changed
       (_ni_read_cache may have skipped some bytes) */
    
    if (fs->aio_read_pos < fs->pos_in_file)
    {						/* Fix if skipped bytes */
      if (fs->aio_read_pos + read_length < fs->pos_in_file)
      {
	read_length= 0;				/* Skipp block */
	next_pos_in_file= fs->pos_in_file;
      }
      else
      {
	my_off_t offset= (fs->pos_in_file - fs->aio_read_pos);
	fs->pos_in_file= fs->aio_read_pos; /* Whe are here */
	fs->rc_pos= fs->rc_request_pos+offset;
	read_length-= offset;			/* Bytes left from rc_pos */
      }
    }
#ifndef DBUG_OFF
    if (fs->aio_read_pos > fs->pos_in_file)
    {
      my_errno= EINVAL;
      DBUG_RETURN(fs->read_length= -1);
    }
#endif
    /* Copy found bytes to buffer */
    length= min(Count,read_length);
    memcpy(Buffer,fs->rc_pos,(my_size_t) length);
    Buffer+= length;
    Count-= length;
    left_length+= length;
    fs->rc_end= fs->rc_pos + read_length;
    fs->rc_pos+= length;
  }
  else
    next_pos_in_file= (fs->pos_in_file+ (uint)
		       (fs->rc_end - fs->rc_request_pos));
  
  /* If reading large blocks, or first read or read with skipp */
  if (Count)
  {
    if (next_pos_in_file == fs->end_of_file)
    {
      fs->error=(int) (read_length+left_length);
      DBUG_RETURN(1);
    }
    VOID(my_seek(fs->file,next_pos_in_file,MY_SEEK_SET,MYF(0)));
    read_length= IO_SIZE*2- (uint) (next_pos_in_file & (IO_SIZE-1));
    if (Count < read_length)
    {					/* Small block, read to cache */
      if ((read_length= my_read(fs->file,fs->rc_request_pos,
				read_length, fs->myflags)) == (uint) -1)
	DBUG_RETURN(fs->error= -1);
      use_length= min(Count,read_length);
      memcpy(Buffer,fs->rc_request_pos,(my_size_t) use_length);
      fs->rc_pos= fs->rc_request_pos+Count;
      fs->rc_end= fs->rc_request_pos+read_length;
      fs->pos_in_file= next_pos_in_file;	/* Start of block in cache */
      next_pos_in_file+= read_length;
      
      if (Count != use_length)
      {					/* Didn't find hole block */
	if (fs->myflags & (MY_WME | MY_FAE | MY_FNABP) && Count != org_Count)
	  my_error(EE_EOFERR, MYF(ME_BELL+ME_WAITTANG),
		   my_filename(fs->file),my_errno);
	fs->error= (int) (read_length+left_length);
	DBUG_RETURN(1);
      }
    }
    else
    {						/* Big block, don't cache it */
      if ((read_length= my_read(fs->file,Buffer,(uint) Count,fs->myflags))
	  != Count)
      {
	fs->error= read_length == (uint)  -1 ? -1 : read_length+left_length;
	DBUG_RETURB(1);
      }
      fs->rc_pos= fs->rc_end= fs->rc_request_pos;
      fs->pos_in_file= (next_pos_in_file+=Count);
    }
  }
  
  /* Read next block with asyncronic io */
  max_length= fs->end_of_file - next_pos_in_file;
  diff_length= (next_pos_in_file & (IO_SIZE-1));
  
  if (max_length > (my_off_t) fs->read_length - diff_length)
    max_length= (my_off_t) fs->read_length - diff_length;
  if (fs->rc_request_pos != fs->buffer)
    read_buffer= fs->buffer;
  else
    read_buffer= fs->buffer+fs->read_length;
  fs->aio_read_pos= next_pos_in_file;
  if (max_length)
  {
    fs->aio_result.result.aio_errno= AIO_INPROGRESS;	/* Marker for test */
    DBUG_PRINT("aioread", ("filepos: %ld  length: %ld",
			   (ulong) next_pos_in_file,(ulong) max_length));
    if (aioread(fs->file, read_buffer, (int) max_length,
		(my_off_t) next_pos_in_file, MY_SEEK_SET,
		&fs->aio_result.result))
    {						/* Skipp async io */
      my_errno=errno;
      DBUG_PRINT("error",("got error: %d, aio_result: %d from aioread, async skipped",
			  errno, fs->aio_result.result.aio_errno));
      if (info->rc_request_pos != info->buffer)
      {
	bmove(info->buffer,info->rc_request_pos,
	      (uint) (info->rc_end - info->rc_pos));
	info->rc_request_pos= info->buffer;
	info->rc_pos-= info->read_length;
	info->rc_end-= info->read_length;
      }
      info->read_length= info->buffer_length;	/* Use hole buffer */
      info->read_function= _my_b_read;		/* Use normal IO_READ next */
    }
    else
      info->inited= info->aio_result.pending= 1;
  }
  DBUG_RETURN(0);					/* Block read, async in use */
}
#endif

my_bool u_real_open_cached_file(MY_FILE *fs)
{
  char name_buff[FN_REFLEN];
  MY_FILENAME *fname= &fs->name;
  char *old_buffer= fname->buffer;
  char *old_base= fname->base ? fname->base : 
    !old_buffer ? 0 : 
    fname->buffer[fname->buf_length-1]==FN_LIBCHAR ? 
    fname->buffer+fname->buf_length : fname->buffer+fname->buf_length+1;
  int len;
  
  DBUG_ENTER("u_real_open_cached_file");

  if (fname->base && fname->buffer!=fname->base)
    *(fname->base-1)= 0;
  if ((fs->file_descr= 
       create_temp_file(name_buff, 
			fname->buffer==fname->base ? 0 : fname->buffer, 
			fname->base, 
			(O_RDWR|O_BINARY|O_TRUNC|O_TEMPORARY|O_SHORT_LIVED),
			MYF(MY_WME))) < 0)
    DBUG_RETURN(1);
  
  len= strlen(name_buff);

  if (!(fname->buffer= my_realloc(fname->buffer,len+1,MYF(0))))
    DBUG_RETURN(1);

  
  my_free(fname->buffer,MYF(0));
  if (!(fname->buffer= my_memdup(name_buff,len+1,MYF(0))))
    DBUG_RETURN(1);
  
  fname->alloced_length= len;
  fname->buf_length= len;
  
#ifdef FN_DEVCHAR
  fname->dir= fname->buffer + (fname->dir - old_buffer);
#else
  fname->dir= fname->buffer;
#endif
  
  if (old_base)
    fname->base= fname->buffer+(old_base-old_buffer);
  
  DBUG_RETURN(0);
}

/*
  Open tempfile cached by IO_CACHE
  Should be used when no seeks are done (only reinit_io_buff)
  Return 0 if cache is inited ok
  The actual file is created when the IO_CACHE buffer gets filled
  If dir is not given, use TMPDIR.
*/
MY_FILE *u_open_cached_file(MY_FILE *res, const char* dir, 
			    const char *prefix, uint cache_size, myf myflags)
{
  DBUG_ENTER("u_open_cached_file");
  
  if (myflags & USE_ALLOCATED_MY_FILE || !res)
  {
    if (!(res= (MY_FILE* )my_malloc(sizeof(MY_FILE), MYF(0))))
      DBUG_RETURN(0);                            /* Not enough memory */
    myflags|= USE_ALLOCATED_MY_FILE;
  }
  bzero((char*) res, sizeof(MY_FILE));
  res->myflags= myflags;
  
  if(init_tmp_file_template(&res->name,dir,prefix))
  {
    free_my_file(res);
    DBUG_RETURN(0);
  }
  
  res->cache.buffer= 0;				/* Mark that not open */
  if (init_u_cache(res,cache_size,EN_WRITE_CACHE,0L,0,
		   MYF(myflags | MY_NABP)))
  {
    free_my_file(res);
    DBUG_RETURN(0);
  }
  
  res->file_descr= -1;
  
  DBUG_RETURN(res);
}


/* Flush write cache */

#ifdef THREAD

#define lock_append_buffer(info) \
 pthread_mutex_lock(&(info)->append_buffer_lock)

#define unlock_append_buffer(info) \
 pthread_mutex_unlock(&(info)->append_buffer_lock)

#define LOCK_APPEND_BUFFER if (need_append_buffer_lock) \
  lock_append_buffer(info);

#define UNLOCK_APPEND_BUFFER if (need_append_buffer_lock) \
  unlock_append_buffer(info);

#else
#define lock_append_buffer(info)
#define unlock_append_buffer(info)
#define LOCK_APPEND_BUFFER
#define UNLOCK_APPEND_BUFFER
#endif

int _flush_u_cache(MY_FILE *file, int need_append_buffer_lock)
{
  MY_FILE_CACHE *info= &file->cache;
  uint length;
  my_bool append_cache;
  my_off_t pos_in_file;
  DBUG_ENTER("_flush_u_cache");
  DBUG_PRINT("info",("ptr=%p, file=%d",file,file->file_descr));

  if (!(append_cache = (info->type == EN_SEQ_READ_APPEND)))
    need_append_buffer_lock=0;
  
  if (info->type == EN_WRITE_CACHE || append_cache)
  {
    if (file->file_descr == -1)
    {
      if (u_real_open_cached_file(file))
	DBUG_RETURN((info->error= -1));
    }
    LOCK_APPEND_BUFFER;
    
    if ((length=(uint) (info->write_pos - info->write_buffer)))
    {
      pos_in_file=file->pos_in_file;
      /* if we have append cache, we always open the file with
	 O_APPEND which moves the pos to EOF automatically on every write
      */
      if (!append_cache && info->seek_not_done)
      {					/* File touched, do seek */
	if (my_u_super_seek(file,pos_in_file,MY_SEEK_SET,MYF(0)) ==
	    MY_FILEPOS_ERROR)
	{
	  UNLOCK_APPEND_BUFFER;
	  DBUG_RETURN((info->error= -1));
	}
	if (!append_cache)
	  info->seek_not_done=0;
      }
      
      info->write_end= (info->write_buffer+info->buffer_length-
			((pos_in_file+length) & (IO_SIZE-1)));
      
      if (my_u_super_write(file,info->write_buffer,length,
			   file->myflags | MY_NABP))
	info->error= -1;
      else
	info->error= 0;
      
      if (!append_cache)
	file->pos_in_file+= length;	
      
      if (!append_cache)
      {
        set_if_bigger(file->end_of_file,(pos_in_file+length));
      }
      else
      {
	file->end_of_file+=(info->write_pos-info->append_read_pos);
	DBUG_ASSERT(file->end_of_file == my_u_super_tell(file,MYF(0)));
      }
      
      info->append_read_pos=info->write_pos=info->write_buffer;
      UNLOCK_APPEND_BUFFER;
      DBUG_RETURN(info->error);
    }
  }
#ifdef HAVE_AIOWAIT
  else if (info->type != READ_NET)
  {
    my_aiowait(&info->aio_result);		/* Wait for outstanding req */
    info->inited=0;
  }
#endif
  UNLOCK_APPEND_BUFFER;
  DBUG_RETURN(0);
}


/* Returns != 0 if error on write */
my_size_t my_cache_write(MY_FILE *file, const byte *Buffer, my_size_t Count, 
			 myf MyFlags __attribute__((unused)))
{
  MY_FILE_CACHE *info= &file->cache;
  uint rest_length,length;
  my_size_t old_Count= Count;
  
  DBUG_ENTER("my_cache_write");
  
  if (file->pos_in_file+info->buffer_length > file->end_of_file)
  {
    my_errno=errno=EFBIG;
    DBUG_RETURN(info->error = -1);
  }
  
  rest_length=(uint) (info->write_end - info->write_pos);
  memcpy(info->write_pos,Buffer,(my_size_t) rest_length);
  Buffer+=rest_length;
  Count-=rest_length;
  info->write_pos+=rest_length;
  if (flush_u_cache(file))
    DBUG_RETURN(1);
  if (Count >= IO_SIZE)
  {					/* Fill first intern buffer */
    length=Count & (uint) ~(IO_SIZE-1);
    if (info->seek_not_done)
    {					/* File touched, do seek */
      VOID(my_u_super_seek(file,file->pos_in_file,MY_SEEK_SET,MYF(0)));
      info->seek_not_done=0;
    }
    if (my_u_super_write(file,Buffer, length, file->myflags | MY_NABP))
      DBUG_RETURN(info->error= -1);
    Count-=length;
    Buffer+=length;
    file->pos_in_file+=length;
  }
  memcpy(info->write_pos,Buffer,(my_size_t) Count);
  info->write_pos+=Count;
  DBUG_RETURN(old_Count);
}


my_size_t my_cache_pwrite(MY_FILE *fs, const byte *a, my_size_t b, 
			  my_off_t c, myf MyFlags)
{
  DBUG_ENTER("my_cache_pwrite");
  my_cache_seek(fs, c, SEEK_SET, MyFlags);
  DBUG_RETURN(my_cache_write(fs, a, b, MyFlags));
}


int end_u_cache(MY_FILE *file)
{
  MY_FILE_CACHE *info= &file->cache;
  int error=0;
  EN_IO_CACHE_CALLBACK pre_close;
  DBUG_ENTER("end_u_cache");
  
  if (!info->buffer)
    DBUG_RETURN(0);
  
  if ((pre_close=info->pre_close))
    (*pre_close)(file);
  if (info->alloced_buffer)
  {
    info->alloced_buffer=0;
    if (file->file_descr!=-1)			/* File doesn't exist */
      error=flush_u_cache(file);
    my_free((gptr) info->buffer,MYF(MY_WME));
    info->buffer=info->read_pos=info->read_end=(byte*) 0;
  }
  if (info->type == EN_SEQ_READ_APPEND)
  {
    /* Destroy allocated mutex */
    info->type= EN_UNDEF_CACHE;
#ifdef THREAD
    pthread_mutex_destroy(&info->append_buffer_lock);
#endif
  }
  if(file->vmt->parent_vmt)
    file->vmt = file->vmt->parent_vmt;
  DBUG_RETURN(error);
} /* end_io_cache */


int my_cache_close(MY_FILE *file, myf MyFlags)
{
  MY_FILE_CACHE *cache= &file->cache;
  int res;
  DBUG_ENTER("my_cache_close");
  if ( !cache->buffer )
  {
    if (file->file_descr >= 0)
      res = my_u_super_close(file, MyFlags);
  }
  else
  {
    (void) end_u_cache(file);
    if (file->file_descr >= 0)
      res = my_u_close(file, MyFlags);
  }
  DBUG_RETURN(res);
}


int my_cache_chsize(MY_FILE *fs, my_off_t newlength, int filler, myf MyFlags)
{
  int cres = my_u_super_chsize(fs,newlength,filler,MyFlags);
  DBUG_ENTER("my_cache_close");
  if (cres) 
    DBUG_RETURN(cres);
  /*
    
  TODO: truncate buffer..
  
  */
  DBUG_RETURN(0);
}


my_off_t my_cache_tell(MY_FILE *file, myf MyFlags __attribute__((unused)))
{
  MY_FILE_CACHE *info= &file->cache;
  DBUG_ENTER("my_cache_tell");
  DBUG_RETURN((file)->pos_in_file + 
	      (uint) (*(info)->current_pos - (info)->request_pos));
}

my_off_t my_cache_append_tell(MY_FILE *info, myf MyFlags __attribute__((unused)))
{
  /*
    Prevent optimizer from putting res in a register when debugging
    we need this to be able to see the value of res when the assert fails
  */
  dbug_volatile my_off_t res; 

  /*
    We need to lock the append buffer mutex to keep flush_io_cache()
    from messing with the variables that we need in order to provide the
    answer to the question.
  */
#ifdef THREAD
  pthread_mutex_lock(&info->cache.append_buffer_lock);
#endif
#ifndef DBUG_OFF
  /*
    Make sure EOF is where we think it is. Note that we cannot just use
    my_tell() because we have a reader thread that could have left the
    file offset in a non-EOF location
  */
  {
    volatile my_off_t save_pos;
    save_pos = my_u_super_tell(info,MYF(0));
    my_u_super_seek(info,(my_off_t)0,MY_SEEK_END,MYF(0));
    /*
      Save the value of my_tell in res so we can see it when studying coredump
    */
    DBUG_ASSERT(info->end_of_file - (info->cache.append_read_pos-info->cache.write_buffer)
		== (res=my_u_super_tell(info,MYF(0))));
    my_u_super_seek(info,save_pos,MY_SEEK_SET,MYF(0));
  }
#endif  
  res = info->end_of_file + (info->cache.write_pos-info->cache.append_read_pos);
#ifdef THREAD
  pthread_mutex_unlock(&info->cache.append_buffer_lock);
#endif
  return res;
}


/*
  Make next read happen at the given position
  For write cache, make next write happen at the given position
*/
my_off_t my_cache_seek(MY_FILE *file, my_off_t pos, 
		       int whence __attribute__((unused)), 
		       myf MyFlags __attribute__((unused)))
{
  MY_FILE_CACHE *info= &file->cache;
  my_off_t res = file->pos_in_file;
  my_off_t offset;  
  DBUG_ENTER("my_cache_seek");
  DBUG_PRINT("enter",("pos: %lu", (ulong) pos));
  
  /*
    TODO: verify that it is OK to do seek in the non-append
    area in SEQ_READ_APPEND cache
  */
  /* TODO:
     a) see if this always works
     b) see if there is a better way to make it work
  */
  if (info->type == EN_SEQ_READ_APPEND)
    flush_u_cache(file);
  
  offset= (pos - file->pos_in_file);
  
  if (info->type == EN_READ_CACHE || info->type == EN_SEQ_READ_APPEND)
  {
    /* TODO: explain why this works if pos < info->pos_in_file */
    if ((ulonglong) offset < (ulonglong) (info->read_end - info->buffer))
    {
      /* The read is in the current buffer; Reuse it */
      info->read_pos = info->buffer + offset;
      DBUG_RETURN(res);
    }
    else
    {
      /* Force a new read on next my_b_read */
      info->read_pos=info->read_end=info->buffer;
    }
  }
  else if (info->type == EN_WRITE_CACHE)
  {
    /* If write is in current buffer, reuse it */
    if ((ulonglong) offset <
	(ulonglong) (info->write_end - info->write_buffer))
    {
      info->write_pos = info->write_buffer + offset;
      DBUG_RETURN(res);
    }
    flush_u_cache(file);
    /* Correct buffer end so that we write in increments of IO_SIZE */
    info->write_end=(info->write_buffer+info->buffer_length-
		     (pos & (IO_SIZE-1)));
  }
  file->pos_in_file=pos;
  info->seek_not_done=1;
  DBUG_RETURN(res);
}


/*
  Do sequential read from the SEQ_READ_APPEND cache
  we do this in three stages:
  - first read from info->buffer
  - then if there are still data to read, try the file descriptor
  - afterwards, if there are still data to read, try append buffer
*/
my_size_t my_cache_seq_read(MY_FILE *file, byte *Buffer, my_size_t Count,
			    myf MyFlags __attribute__((unused)))
{
  MY_FILE_CACHE *info= &file->cache;
  uint length,diff_length,left_length,save_count= Count;
  my_off_t max_length, pos_in_file;

  DBUG_ENTER("my_cache_seq_read");
  DBUG_PRINT("info",("fd=%d ptr=%p Count=%d",
		     (int)file->file_descr,file,(int)Count));
  
  /* first, read the regular buffer */
  if ((left_length=(uint) (info->read_end-info->read_pos)))
  {
    DBUG_ASSERT(Count > left_length);	/* User is not using my_b_read() */
    memcpy(Buffer,info->read_pos, (my_size_t) (left_length));
    Buffer+=left_length;
    Count-=left_length;
  }
  lock_append_buffer(info);
  
  /* pos_in_file always point on where info->buffer was read */
  if ((pos_in_file=file->pos_in_file+(uint) (info->read_end - info->buffer)) >=
      file->end_of_file)
    goto read_append_buffer;
  
  /*
    With read-append cache we must always do a seek before we read,
    because the write could have moved the file pointer astray
  */
  VOID(my_u_super_seek(file,pos_in_file,MY_SEEK_SET,MYF(0)));
  info->seek_not_done=0;
  
  diff_length=(uint) (pos_in_file & (IO_SIZE-1));
  
  /* now the second stage begins - read from file descriptor */
  if (Count >= (uint) (IO_SIZE+(IO_SIZE-diff_length)))
  {					/* Fill first intern buffer */
    uint read_length;
    
    length=(Count & (uint) ~(IO_SIZE-1))-diff_length;
    if ((read_length= my_u_super_read(file,Buffer,length,file->myflags))!=
	length)
    {
      info->error= -1;
      unlock_append_buffer(info);
      DBUG_RETURN(1);
    }
    Count-=read_length;
    Buffer+=read_length;
    pos_in_file+=read_length;
    
    if (read_length != (uint) length)
    {
      /*
	We only got part of data;  Read the rest of the data from the
	write buffer
      */
      goto read_append_buffer;
    }
    left_length+=length;
    diff_length=0;
  }
  
  max_length=info->read_length-diff_length;
  if (max_length > (file->end_of_file - pos_in_file))
    max_length = file->end_of_file - pos_in_file;
  if (!max_length)
  {
    if (Count)
      goto read_append_buffer;
    length=0;				/* Didn't read any more chars */
  }
  else
  {
    length= my_u_super_read(file,info->buffer,(uint) max_length,
			    file->myflags);
    if (length == (uint) -1)
    {
      info->error= -1;
      unlock_append_buffer(info);
      DBUG_RETURN(1);
    }
    if (length < Count)
    {
      memcpy(Buffer,info->buffer,(my_size_t) length);
      Count -= length;
      Buffer += length;
      
      /*
	added the line below to make
	DBUG_ASSERT(pos_in_file==info->end_of_file) pass.
	otherwise this does not appear to be needed
      */
      pos_in_file += length;
      goto read_append_buffer;
    }
  }
  unlock_append_buffer(info);
  info->read_pos=info->buffer+Count;
  info->read_end=info->buffer+length;
  file->pos_in_file=pos_in_file;
  memcpy(Buffer,info->buffer,(my_size_t) Count);
  DBUG_RETURN(save_count);
  
read_append_buffer:
  
  /*
    Read data from the current write buffer.
    Count should never be == 0 here (The code will work even if count is 0)
  */
  
  {
    /* First copy the data to Count */
    uint len_in_buff = (uint) (info->write_pos - info->append_read_pos);
    uint copy_len;
    uint transfer_len;
    
    DBUG_ASSERT(info->append_read_pos <= info->write_pos);
    /*
      TODO: figure out if the assert below is needed or correct.
    */
    DBUG_ASSERT(pos_in_file == file->end_of_file); 
    copy_len=min(Count, len_in_buff);
    memcpy(Buffer, info->append_read_pos, copy_len);
    info->append_read_pos += copy_len;
    Count -= copy_len;
    if (Count)
      info->error = save_count - Count;
    
    /* Fill read buffer with data from write buffer */
    memcpy(info->buffer, info->append_read_pos,
	   (my_size_t) (transfer_len=len_in_buff - copy_len));
    info->read_pos= info->buffer;
    info->read_end= info->buffer+transfer_len;
    info->append_read_pos=info->write_pos;
    file->pos_in_file=pos_in_file+copy_len;
    file->end_of_file+=len_in_buff;
  }
  unlock_append_buffer(info);
  DBUG_RETURN(Count ? 1 : save_count);
}


/*
  Use this to reset cache to re-start reading or to change the type
  between READ_CACHE <-> WRITE_CACHE
  If we are doing a reinit of a cache where we have the start of the file
  in the cache, we are reusing this memory without flushing it to disk.
*/
my_bool reinit_u_cache(MY_FILE *file, enum en_cache_type type,
		       my_off_t seek_offset,
		       pbool use_async_io __attribute__((unused)),
		       pbool clear_cache)
{
  MY_FILE_CACHE *info= &file->cache;
  DBUG_ENTER("reinit_u_cache");
  DBUG_PRINT("enter",("ptr=%p, fd=%d, type: %d  seek_offset: %lu  clear_cache: %d",
		      file, (int)file->file_descr, type, 
		      (ulong) seek_offset, (int) clear_cache));
  
  /* One can't do reinit with the following types */
  DBUG_ASSERT(type != EN_READ_NET && info->type != EN_READ_NET &&
	      type != EN_WRITE_NET && info->type != EN_WRITE_NET &&
	      type != EN_SEQ_READ_APPEND && info->type != EN_SEQ_READ_APPEND);

  /* If the whole file is in memory, avoid flushing to disk */
  if (! clear_cache &&
      seek_offset >= file->pos_in_file &&
      seek_offset <= my_u_tell(file,MYF(0)))
  {
    /* Reuse current buffer without flushing it to disk */
    byte *pos;
    if (info->type == EN_WRITE_CACHE && type == EN_READ_CACHE)
    {
      info->read_end=info->write_pos;
      file->end_of_file=my_u_tell(file,MYF(0));
    }
    else if (type == EN_WRITE_CACHE)
    {
      if (info->type == EN_READ_CACHE)
	info->write_end=info->write_buffer+info->buffer_length;
      file->end_of_file = ~(my_off_t) 0;
    }
    pos=info->request_pos+(seek_offset-file->pos_in_file);
    if (type == EN_WRITE_CACHE)
      info->write_pos=pos;
    else
      info->read_pos= pos;
#ifdef HAVE_AIOWAIT
    my_aiowait(&info->aio_result);		/* Wait for outstanding req */
#endif
  }
  else
  {
    /*
      If we change from WRITE_CACHE to READ_CACHE, assume that everything
      after the current positions should be ignored
    */
    if (info->type == EN_WRITE_CACHE && type == EN_READ_CACHE)
      file->end_of_file=my_u_tell(file,MYF(0));
    /* flush cache if we want to reuse it */
    if (!clear_cache && flush_u_cache(file))
      DBUG_RETURN(1);
    file->pos_in_file=seek_offset;
    /* Better to do always do a seek */
    info->seek_not_done=1;
    info->request_pos=info->read_pos=info->write_pos=info->buffer;
    if (type == EN_READ_CACHE)
    {
      info->read_end=info->buffer;		/* Nothing in cache */
    }
    else
    {
      info->write_end=(info->buffer + info->buffer_length -
		       (seek_offset & (IO_SIZE-1)));
      file->end_of_file= ~(my_off_t) 0;
    }
  }
  info->type=type;
  info->error=0;
  init_u_functions(file,type);
  
#ifdef HAVE_AIOWAIT
  if (use_async_io && ! my_disable_async_io &&
      ((ulong) info->buffer_length <
       (ulong) (info->end_of_file - seek_offset)))
  {
    info->read_length=info->buffer_length/2;
    info->read_function=_my_b_async_read;
  }
  info->inited=0;
#endif
  DBUG_RETURN(0);
} /* reinit_u_cache */

extern int my_u_cache_append(MY_FILE *file, const byte *Buffer, uint Count)
{
  uint rest_length,length;
  MY_FILE_CACHE *info= &file->cache;

  DBUG_ENTER("my_u_cache_append");

  lock_append_buffer(info);
  rest_length=(uint) (info->write_end - info->write_pos);
  DBUG_PRINT("info",("ptr=%p, fd=%d, Count=%d, rest_length=%d",
		     file,(int)file->file_descr,(int)Count,(int)rest_length));
  if (Count <= rest_length)
    goto end;

  memcpy(info->write_pos,Buffer,(size_t) rest_length);
  Buffer+=rest_length;
  Count-=rest_length;
  info->write_pos+=rest_length;
  if (_flush_u_cache(file,0))
  {
    unlock_append_buffer(info);
    DBUG_RETURN(1);
  }
  if (Count >= IO_SIZE)
  {					/* Fill first intern buffer */
    length=Count & (uint) ~(IO_SIZE-1);
    if (my_u_super_write(file, Buffer, (uint) length,file->myflags | MY_NABP))
    {
      unlock_append_buffer(info);
     DBUG_RETURN(info->error= -1);
    }
    Count-=length;
    Buffer+=length;
    file->end_of_file+=length;
  }

end:
  memcpy(info->write_pos,Buffer,(size_t) Count);
  info->write_pos+=Count;
  unlock_append_buffer(info);
  DBUG_RETURN(0);
}

extern int my_u_safe_write(MY_FILE *file, const byte *buf, uint len)
{
  /*
    Sasha: We are not writing this with the ? operator to avoid hitting
    a possible compiler bug. At least gcc 2.95 cannot deal with 
    several layers of ternary operators that evaluated comma(,) operator
    expressions inside - I do have a test case if somebody wants it
  */
  if (file->cache.type == EN_SEQ_READ_APPEND)
    return my_u_cache_append(file, buf,len);
  return my_u_write(file, buf, len, MYF(0));
}

#ifdef THREAD

#define IO_ROUND_UP(X) (((X)+IO_SIZE-1) & ~(IO_SIZE-1))
#define IO_ROUND_DN(X) ( (X)            & ~(IO_SIZE-1))

/* Prepare IO_CACHE for shared use */
void init_u_cache_share(MY_FILE *file, MY_CACHE_SHARE *s, uint num_threads)
{
  MY_FILE_CACHE *info= &file->cache;
  DBUG_ASSERT(info->type == EN_READ_CACHE);
  pthread_mutex_init(&s->mutex, MY_MUTEX_INIT_FAST);
  pthread_cond_init (&s->cond, 0);
  s->count=num_threads-1;
  s->active=0; /* to catch errors */
  info->share=s;
  //info->read_function=_my_b_read_r;
}

/*
  Remove a thread from shared access to IO_CACHE
  Every thread should do that on exit for not
  to deadlock other threads
*/
void remove_u_thread(MY_FILE *file)
{
  MY_FILE_CACHE *info= &file->cache;
  pthread_mutex_lock(&info->share->mutex);
  if (! info->share->count--)
    pthread_cond_signal(&info->share->cond);
  pthread_mutex_unlock(&info->share->mutex);
}

static int lock_u_cache(MY_FILE *file)
{
  MY_FILE_CACHE *info= &file->cache;
  pthread_mutex_lock(&info->share->mutex);
  if (!info->share->count)
    return 1;

  --(info->share->count);
  pthread_cond_wait(&info->share->cond, &info->share->mutex);
  /*
    count can be -1 here, if one thread was removed (remove_io_thread)
    while all others were locked (lock_io_cache).
    If this is the case, this thread behaves as if count was 0 from the
    very beginning, that is returns 1 and does not unlock the mutex.
  */
  if (++(info->share->count))
  {
    pthread_mutex_unlock(&info->share->mutex);
    return 0;
  }
  return 1;
}

static void unlock_u_cache(MY_FILE *file)
{
  MY_FILE_CACHE *info= &file->cache;
  pthread_cond_broadcast(&info->share->cond);
  pthread_mutex_unlock(&info->share->mutex);
}


/*
  Read from IO_CACHE when it is shared between several threads.
  It works as follows: when a thread tries to read from a file
  (that is, after using all the data from the (shared) buffer),
  it just hangs on lock_io_cache(), wating for other threads.
  When the very last thread attempts a read, lock_io_cache()
  returns 1, the thread does actual IO and unlock_io_cache(),
  which signals all the waiting threads that data is in the buffer.
*/
size_t my_cache_read_r(MY_FILE *file, byte *Buffer, size_t Count,
					myf MyFlags __attribute__((unused)))
{
  my_off_t pos_in_file;
  uint length,diff_length,read_len;
  MY_FILE_CACHE *info= &file->cache;
  DBUG_ENTER("_my_b_read_r");

  if ((read_len=(uint) (info->read_end-info->read_pos)))
  {
    DBUG_ASSERT(Count >= read_len);	/* User is not using my_b_read() */
    memcpy(Buffer,info->read_pos, (size_t) (read_len));
    Buffer+=read_len;
    Count-=read_len;
  }
  while (Count)
  {
    int cnt, len;

    pos_in_file= file->pos_in_file + (uint)(info->read_end - info->buffer);
    diff_length= (uint) (pos_in_file & (IO_SIZE-1));
    length=IO_ROUND_UP(Count+diff_length)-diff_length;
    length=(length <= info->read_length) ?
                   length + IO_ROUND_DN(info->read_length - length) :
                   length - IO_ROUND_UP(length - info->read_length) ;
    if (info->type != READ_FIFO &&
	(length > (uint) (file->end_of_file - pos_in_file)))
      length= (uint) (file->end_of_file - pos_in_file);
    if (length == 0)
    {
      info->error=(int) read_len;
      DBUG_RETURN(1);
    }
    if (lock_u_cache(file))
    {
      info->share->active=file;
      if (info->seek_not_done)             /* File touched, do seek */
        VOID(my_u_seek(file,pos_in_file,MY_SEEK_SET,MYF(0)));
      len=(int)my_u_read(file,info->buffer, length, file->myflags);
      info->read_end=info->buffer + (len == -1 ? 0 : len);
      info->error=(len == (int)length ? 0 : len);
      file->pos_in_file=pos_in_file;
      unlock_u_cache(file);
    }
    else
    {
      info->error=          info->share->active->cache.error;
      info->read_end=       info->share->active->cache.read_end;
      file->pos_in_file=    info->share->active->pos_in_file;
      len= (info->error == -1 ? -1 : info->read_end-info->buffer);
    }
    info->read_pos=info->buffer;
    info->seek_not_done=0;
    if (len <= 0)
    {
      info->error=(int) read_len;
      DBUG_RETURN(1);
    }
    cnt=((uint) len > Count) ? (int) Count : len;
    memcpy(Buffer,info->read_pos, (size_t)cnt);
    Count -=cnt;
    Buffer+=cnt;
    read_len+=cnt;
    info->read_pos+=cnt;
  }
  DBUG_RETURN(0);
}

#endif
