/*
   Copyright (C) 2000 MySQL AB & MySQL Finland AB & TCX DataKonsult AB
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <my_global.h>
#include <my_dir.h>
#include <my_alarm.h>
#include <my_sys.h>
#include <mysys_err.h>
#include <errno.h>
#include <m_string.h>

/* VMT for simple files.. */
MY_FILE_VMT simple_file_vmt = 
{
  0,
  &my_simple_read,
  &my_simple_pread,
  &my_simple_write,
  &my_simple_pwrite,
  &my_simple_close,
  &my_simple_lock,
  &my_simple_chsize,
  &my_simple_fstat,
  &my_simple_tell,
  &my_simple_seek,
  &my_simple_free
};

void my_simple_free(MY_FILE *my_file)
{
}

bool my_simple_init_my_file(MY_FILE * my_file, enum en_file_type type_of_file,
			    uint error_message_number, myf MyFlags)
{
  DBUG_ENTER("my_simple_init_my_file");
  
  if ((int) my_file->file_descr >= 0)
  {
    my_file->file_type = type_of_file;
    my_file->vmt= &simple_file_vmt;
    my_file->cache.buffer= my_file->cache.write_pos= my_file->cache.write_end=
			   my_file->cache.read_pos= my_file->cache.read_end= 0;
#if defined(THREAD) && !defined(HAVE_PREAD)
    pthread_mutex_init(&fs->mutex, MY_MUTEX_INIT_FAST);
#endif
    DBUG_RETURN(0);
  }
  
  my_errno=errno;
  DBUG_PRINT("error",("Got error %d on open",my_errno));
  if (MyFlags & (MY_FFNF | MY_FAE | MY_WME))
    my_error(error_message_number, MYF(ME_BELL+ME_WAITTANG),
	     my_file->name.buffer , my_errno);
  DBUG_RETURN(1);
}

bool my_simple_open(MY_FILE * my_file, int Flags, myf MyFlags)
{
  char file_name[FN_REFLEN];
  char *file_for_open;
  
  DBUG_ENTER("my_simple_open");
  DBUG_PRINT("my",("ptr=%p, Flags=%d, MyFlags=%d", my_file, Flags, MyFlags));
  
  file_for_open= (char *)system_my_filename(file_name, my_file);
  
#if defined(MSDOS) || defined(__WIN__) || defined(__EMX__) || defined(OS2)
  if (Flags & O_SHARE)
    my_file->file_descr = sopen((my_string) file_for_open,
				(Flags & ~O_SHARE) | O_BINARY, 
				SH_DENYNO, MY_S_IREAD | MY_S_IWRITE);
  else
    my_file->file_descr = open((my_string) file_for_open, Flags | O_BINARY,
			       MY_S_IREAD | MY_S_IWRITE);
#elif !defined(NO_OPEN_3)
  my_file->file_descr = open(file_for_open, Flags, my_umask);  /* Normal unix */
#else
  my_file->file_descr = open((my_string) file_for_open, Flags);
#endif

  DBUG_PRINT("my",("fd=%d",my_file->file_descr));
  
  DBUG_RETURN( my_simple_init_my_file(my_file, EN_FILE_BY_OPEN,
				      EE_FILENOTFOUND, MyFlags) );
}

bool my_simple_create(MY_FILE * my_file, int CreateFlags,
		      int access_flags, myf MyFlags)
{
  char file_name[FN_REFLEN];
  char *file_for_create;
  
  DBUG_ENTER("my_simple_create");
  DBUG_PRINT("my",("Name: '%s' CreateFlags: %d  AccessFlags: %d  MyFlags: %d",
                   my_file->name.buffer, CreateFlags, access_flags, MyFlags));

  file_for_create= (char *)system_my_filename(file_name, my_file);
  
#if !defined(NO_OPEN_3) && !defined(__EMX__)
  my_file->file_descr = open((my_string) file_for_create, access_flags | O_CREAT,
			     CreateFlags ? CreateFlags : my_umask);
#elif defined(VMS)
  my_file->file_descr = open((my_string) file_for_create, access_flags | O_CREAT, 0,
			     "ctx=stm","ctx=bin");
#elif defined(MSDOS) || defined(__WIN__) || defined(__EMX__) || defined(OS2)
  if (access_flags & O_SHARE)
    my_file->file_descr = sopen((my_string) file_for_create,
				access_flags | O_CREAT | O_BINARY, 
				SH_DENYNO, MY_S_IREAD | MY_S_IWRITE);  
  else
    my_file->file_descr =  open((my_string) file_for_create,
				access_flags | O_CREAT | O_BINARY, 
				MY_S_IREAD | MY_S_IWRITE);
#else
  my_file->file_descr = open(file_for_create, access_flags);
#endif

  DBUG_PRINT("my",("fd=%d",my_file->file_descr));
  
  DBUG_RETURN(  my_simple_init_my_file(my_file, EN_FILE_BY_CREATE,
				       EE_CANTCREATEFILE, MyFlags));
}

my_size_t my_simple_read(MY_FILE * fs, byte *Buffer, my_size_t Count, myf MyFlags)
{
  uint readbytes,save_count;
  int Filedes= fs->file_descr;
  DBUG_ENTER("my_simple_read");
  DBUG_PRINT("my",("Fd: %d this:%p  Buffer: %lx  Count: %u  MyFlags: %d",
		   Filedes, fs, Buffer, Count, MyFlags));    
  
  save_count= Count;
  for (;;)
  {
    errno= 0;					/* Linux doesn't reset this */
    if ((readbytes = (uint) read(Filedes, Buffer, Count)) != Count)
    {
      my_errno= errno ? errno : -1;
      DBUG_PRINT("warning",("Read only %ld bytes off %ld from %d, errno: %d",
			    readbytes,Count,Filedes,my_errno));
#ifdef THREAD
      if (readbytes == 0 && errno == EINTR)
	continue;				/* Interrupted */
#endif
      if (MyFlags & (MY_WME | MY_FAE | MY_FNABP))
      {
	if ((int) readbytes == -1)
	  my_error(EE_READ, MYF(ME_BELL+ME_WAITTANG),
		   fs->name.buffer,my_errno);
	else if (MyFlags & (MY_NABP | MY_FNABP))
	  my_error(EE_EOFERR, MYF(ME_BELL+ME_WAITTANG),
		   fs->name.buffer,my_errno);
      }
      if ((int) readbytes == -1 ||
	  ((MyFlags & (MY_FNABP | MY_NABP)) && !(MyFlags & MY_FULL_IO)))
	DBUG_RETURN(MY_FILE_ERROR);	/* Return with error */
      if (readbytes > 0 && (MyFlags & MY_FULL_IO))
      {
	Buffer+= readbytes;
	Count-= readbytes;
	continue;
      }
    }
    
    if (MyFlags & (MY_NABP | MY_FNABP) && readbytes!=save_count)
      readbytes= 0;			/* Ok on read */
    else if (MyFlags & MY_FULL_IO)
      readbytes= save_count;
    break;
  }
  DBUG_RETURN(readbytes);
}

my_size_t my_simple_pread(MY_FILE *fs, byte *Buffer, my_size_t Count, 
			  my_off_t offset, myf MyFlags)
{
  uint readbytes;
  int Filedes= fs->file_descr;
  int error;
  
  DBUG_ENTER("my_simple_pread");
  DBUG_PRINT("my",("Fd: %d  Seek: %lu  Buffer: %lx  Count: %u  MyFlags: %d",
		   Filedes, (ulong) offset, Buffer, Count, MyFlags));
  
  for (;;)
  {
#ifndef __WIN__
    errno= 0;					/* Linux doesn't reset this */
#endif
#ifndef HAVE_PREAD
    pthread_mutex_lock(&fs->mutex);
    fs->cache.seek_not_done = 1;
    readbytes= (uint) -1;
    error= (lseek(Filedes, offset, MY_SEEK_SET) == -1L ||
	    (readbytes= (uint) read(Filedes, Buffer, Count)) != Count);
    pthread_mutex_unlock(&fs->mutex);
#else
    error=((readbytes= (uint) pread(Filedes, Buffer, Count, offset)) != Count);
#endif
    if (error)
    {
      my_errno= errno;
      DBUG_PRINT("warning",("Read only %ld bytes off %ld from %d, errno: %d",
			    readbytes,Count,Filedes,my_errno));
#ifdef THREAD
      if (readbytes == 0 && errno == EINTR)
	continue;				/* Interrupted */
#endif
      if (MyFlags & (MY_WME | MY_FAE | MY_FNABP))
      {
	if ((int) readbytes == -1)
	  my_error(EE_READ, MYF(ME_BELL+ME_WAITTANG),
		   fs->name.buffer,my_errno);
	else if (MyFlags & (MY_NABP | MY_FNABP))
	  my_error(EE_EOFERR, MYF(ME_BELL+ME_WAITTANG),
		   fs->name.buffer,my_errno);
      }
      if ((int) readbytes == -1 || (MyFlags & (MY_FNABP | MY_NABP)))
	DBUG_RETURN(MY_FILE_ERROR);		/* Return with error */
    }
    if (MyFlags & (MY_NABP | MY_FNABP))
      DBUG_RETURN(0);				/* Read went ok; Return 0 */
    DBUG_RETURN(readbytes);
  }
}

my_size_t my_simple_write(MY_FILE *fs, const byte *Buffer, 
			  my_size_t Count, myf MyFlags)
{
  int Filedes= fs->file_descr;
  uint writenbytes,errors;
  ulong written;
  DBUG_ENTER("my_simple_write");
  DBUG_PRINT("my",("Fd: %d  Buffer: %lx  Count: %d  MyFlags: %d",
		   Filedes, Buffer, Count, MyFlags));
  
  errors= 0; written= 0L;
  for (;;)
  {
    if ((writenbytes= (uint) write(Filedes, Buffer, Count)) == Count)
      break;
    if ((int) writenbytes != -1)
    {						/* Safeguard */
      written+= writenbytes;
      Buffer+= writenbytes;
      Count-= writenbytes;
    }
    my_errno= errno;
    DBUG_PRINT("error",("Write only %d bytes, error: %d",
			writenbytes,my_errno));
#ifndef NO_BACKGROUND
#ifdef THREAD
    if (my_thread_var->abort)
      MyFlags&= ~ MY_WAIT_IF_FULL;		/* End if aborted by user */
#endif
    if (my_errno == ENOSPC && (MyFlags & MY_WAIT_IF_FULL) &&
	(uint) writenbytes != (uint) -1)
    {
      if (!(errors++ % MY_WAIT_GIVE_USER_A_MESSAGE))
	my_error(EE_DISK_FULL,MYF(ME_BELL | ME_NOREFRESH),
		 fs->name.buffer);
      VOID(sleep(MY_WAIT_FOR_USER_TO_FIX_PANIC));
      continue;
    }
    if (!writenbytes)
    {
      /* We may come here on an interrupt or if the file quote is exeeded */
      if (my_errno == EINTR)
	continue;
      if (!errors++)				/* Retry once */
      {
	errno=EFBIG;				/* Assume this is the error */
	continue;
      }
    }
    else if ((uint) writenbytes != (uint) -1)
      continue;					/* Retry */
#endif
    if (MyFlags & (MY_NABP | MY_FNABP))
    {
      if (MyFlags & (MY_WME | MY_FAE | MY_FNABP))
      {
	my_error(EE_WRITE, MYF(ME_BELL+ME_WAITTANG),
		 fs->name.buffer,my_errno);
      }
      DBUG_RETURN(MY_FILE_ERROR);		/* Error on read */
    }
    else
      break;					/* Return bytes written */
  }
  if (MyFlags & (MY_NABP | MY_FNABP))
    DBUG_RETURN(0);				/* Want only errors */
  DBUG_RETURN(writenbytes+written);
}

my_size_t my_simple_pwrite(MY_FILE *fs, const byte *Buffer, my_size_t Count, 
			   my_off_t offset, myf MyFlags)
{
  int Filedes= fs->file_descr;
  uint writenbytes,errors;
  ulong written;
  DBUG_ENTER("my_simple_pwrite");
  DBUG_PRINT("my",("Fd: %d  Seek: %lu  Buffer: %lx  Count: %d  MyFlags: %d",
		   Filedes, (ulong) offset,Buffer, Count, MyFlags));
  errors= 0; written= 0L;
  
  for (;;)
  {
#ifndef HAVE_PREAD
    int error;
    writenbytes= (uint) -1;
    pthread_mutex_lock(&fs->mutex);
    error= (lseek(Filedes, offset, MY_SEEK_SET) != -1L &&
	     (writenbytes = (uint) write(Filedes, Buffer, Count)) == Count);
    pthread_mutex_unlock(&gs->mutex);
    if (error)
      break;
#else
    if ((writenbytes= (uint) pwrite(Filedes, Buffer, Count,offset)) == Count)
      break;
#endif
    if ((int) writenbytes != -1)
    {					/* Safegueard */
      written+= writenbytes;
      Buffer+= writenbytes;
      Count-= writenbytes;
      offset+= writenbytes;
    }
    my_errno= errno;
    DBUG_PRINT("error",("Write only %d bytes",writenbytes));
#ifndef NO_BACKGROUND
#ifdef THREAD
    if (my_thread_var->abort)
      MyFlags&= ~ MY_WAIT_IF_FULL;		/* End if aborted by user */
#endif
    if (my_errno == ENOSPC && (MyFlags & MY_WAIT_IF_FULL))
    {
      if (!(errors++ % MY_WAIT_GIVE_USER_A_MESSAGE))
	my_error(EE_DISK_FULL,MYF(ME_BELL | ME_NOREFRESH),
		 fs->name.buffer);
      VOID(sleep(MY_WAIT_FOR_USER_TO_FIX_PANIC));
      continue;
    }
    if ((writenbytes == 0 && my_errno == EINTR) ||
	(writenbytes > 0 && (uint) writenbytes != (uint) -1))
      continue;					/* Retry */
#endif
    if (MyFlags & (MY_NABP | MY_FNABP))
    {
      if (MyFlags & (MY_WME | MY_FAE | MY_FNABP))
      {
	my_error(EE_WRITE, MYF(ME_BELL+ME_WAITTANG),
		 fs->name.buffer,my_errno);
      }
      DBUG_RETURN(MY_FILE_ERROR);		/* Error on read */
    }
    else
      break;					/* Return bytes written */
  }
  if (MyFlags & (MY_NABP | MY_FNABP))
    DBUG_RETURN(0);			/* Want only errors */
  DBUG_RETURN(writenbytes+written); /* purecov: inspected */
}

int my_simple_close(MY_FILE *fs, myf MyFlags)
{
  int err;
  DBUG_ENTER("my_simple_close");
  DBUG_PRINT("info",("ptr=%p, fd=%d",fs,fs->file_descr));
  
  if ((err= close(fs->file_descr)))
  {
    DBUG_PRINT("error",("Got error %d on close",err));
    my_errno= errno;
    if (MyFlags & (MY_FAE | MY_WME))
      my_error(EE_BADCLOSE, MYF(ME_BELL+ME_WAITTANG), fs->name.buffer, errno);
  }
  fs->file_descr= -1;
#if defined(THREAD) && !defined(HAVE_PREAD)
  pthread_mutex_destroy(&fs->mutex);
#endif
  fs->file_type= EN_UNOPEN;
  
  DBUG_RETURN(err);
}

int my_simple_lock(MY_FILE * fs, int locktype,
		   my_off_t start, my_off_t length, myf MyFlags)
{
  int fd= fs->file_descr;
#ifdef HAVE_FCNTL
  int value;
  ALARM_VARIABLES;
#endif
  DBUG_ENTER("my_simple_lock");
  DBUG_PRINT("my",("Fd: %d  Op: %d  start: %ld  Length: %ld  MyFlags: %d",
		   fd,locktype,(long) start,(long) length,MyFlags));
  
#ifdef VMS
  DBUG_RETURN(0);
#else
  if (my_disable_locking)
    DBUG_RETURN(0);
  
#if defined(__EMX__) || defined(OS2)
  
  if (!_lock64( fd, locktype, start, length, MyFlags))
    DBUG_RETURN(0);
  
#elif defined(HAVE_LOCKING)
  /* Windows */
  {
    my_bool error;
    pthread_mutex_lock(&fs->mutex);
    if (MyFlags & MY_SEEK_NOT_DONE)
      VOID(my_simple_seek(fs,start,MY_SEEK_SET,MYF(MyFlags & ~MY_SEEK_NOT_DONE)));
    error= locking(fd,locktype,(ulong) length) && errno != EINVAL;
    pthread_mutex_unlock(&fs->mutex);
    if (!error)
      DBUG_RETURN(0);
  }
#else
#if defined(HAVE_FCNTL)
  {
    struct flock lock;
    lock.l_type= (short) locktype;
    lock.l_whence= 0L;
    lock.l_start=(long) start;
    lock.l_len= (long) length;
    if (MyFlags & MY_DONT_WAIT)
    {
      if (fcntl(fd,F_SETLK,&lock) != -1)	/* Check if we can lock */
	DBUG_RETURN(0);			/* Ok, file locked */
      DBUG_PRINT("info",("Was locked, trying with alarm"));
      ALARM_INIT;
      while ((value= fcntl(fd,F_SETLKW,&lock)) && ! ALARM_TEST &&
	     errno == EINTR)
      {			/* Setup again so we don`t miss it */
	ALARM_REINIT;
      }
      ALARM_END;
      if (value != -1)
	DBUG_RETURN(0);
      if (errno == EINTR)
	errno= EAGAIN;
    }
    else if (fcntl(fd,F_SETLKW,&lock) != -1) /* Wait until a lock */
      DBUG_RETURN(0);
  }
#else
  if (MyFlags & MY_SEEK_NOT_DONE)
    VOID(my_simple_seek(fs,start,MY_SEEK_SET,MYF(MyFlags & ~MY_SEEK_NOT_DONE)));
  if (lockf(fd,locktype,length) != -1)
    DBUG_RETURN(0);
#endif /* HAVE_FCNTL */
#endif /* HAVE_LOCKING */
  
  /* We got an error. We don't want EACCES errors */
  my_errno= (errno == EACCES) ? EAGAIN : errno ? errno : -1;
  if (MyFlags & MY_WME)
  {
    if (locktype == F_UNLCK)
      my_error(EE_CANTUNLOCK,MYF(ME_BELL+ME_WAITTANG),my_errno);
    else
      my_error(EE_CANTLOCK,MYF(ME_BELL+ME_WAITTANG),my_errno);
  }
  DBUG_PRINT("error",("my_errno: %d (%d)",my_errno,errno));
  DBUG_RETURN(-1);
#endif	/* ! VMS */
}

/*
  Change size of file.

  SYNOPSIS
    my_simple_chsize()
      my_simple_chsize	File descriptor
      new_length	New file size
      filler		If we don't have truncate, fill up all bytes after
			new_length with this character
      MyFlags		Flags

  DESCRIPTION
    my_simple_chsize() truncates file if shorter, else fill with the filler
    character

  RETURN VALUE
    0	Ok
    1	Error 
*/

int my_simple_chsize(MY_FILE * fs, my_off_t newlength, int filler, myf MyFlags)
{
  int fd= fs->file_descr;
  DBUG_ENTER("my_simple_chsize");
  DBUG_PRINT("my",("fd: %d  length: %lu  MyFlags: %d",fd,(ulong) newlength,
		   MyFlags));
  
#ifdef HAVE_CHSIZE
  if (chsize(fd,(off_t) newlength))
  {
    DBUG_PRINT("error",("errno: %d",errno));
    my_errno= errno;
    if (MyFlags & MY_WME)
      my_error(EE_CANT_CHSIZE,MYF(ME_BELL+ME_WAITTANG),errno);
    DBUG_RETURN(1);
  }
  DBUG_RETURN(0);
#else
  /* if file is shorter, expand with null, else fill unused part with null */
  {
    my_off_t oldsize;
    char buff[IO_SIZE];
    
    oldsize= my_u_seek(fs, 0L, MY_SEEK_END, MYF(MY_WME+MY_FAE));
    DBUG_PRINT("info",("old_size: %ld", (ulong) oldsize));
    
#ifdef HAVE_FTRUNCATE
    if (oldsize > newlength)
    {
      if (ftruncate(fd, (off_t) newlength))
      {
	my_errno= errno;
	DBUG_PRINT("error",("errno: %d",errno));
	if (MyFlags & MY_WME)
	  my_error(EE_CANT_CHSIZE, MYF(ME_BELL+ME_WAITTANG), errno);
	DBUG_RETURN(1);
      }
      DBUG_RETURN(0);
    }
#else
    if (oldsize > newlength)
    {                          /* Fill diff with null */
      VOID(my_u_seek(fd, newlength, MY_SEEK_SET, MYF(MY_WME+MY_FAE)));
      swap(my_off_t, newlength, oldsize);
    }
#endif

    /* Full file with 0 until it's as big as requested */
    bfill(buff, IO_SIZE, filler);
    while (newlength-oldsize > IO_SIZE)
    {
      if (my_u_write(fs,(byte*) buff,IO_SIZE,MYF(MY_NABP)))
	goto err;
      oldsize+= IO_SIZE;
    }
    if (my_u_write(fs,(byte*) buff,newlength-oldsize,MYF(MY_NABP)))
      goto err;
    DBUG_RETURN(0);
err:
    if (MyFlags & MY_WME)
      my_error(EE_CANT_CHSIZE,MYF(ME_BELL+ME_WAITTANG),my_errno);
    DBUG_PRINT("error",("errno: %d",my_errno));
    DBUG_RETURN(1);
  }
#endif
}

int my_simple_fstat(MY_FILE * fs, struct my_stat *stat_area, myf MyFlags)
{
  int Filedes= fs->file_descr;
  DBUG_ENTER("my_simple_fstat");
  DBUG_PRINT("my",("fd: %d MyFlags: %d",Filedes,MyFlags));
  DBUG_RETURN(fstat(Filedes, (struct stat *) stat_area));
}

my_off_t my_simple_tell(MY_FILE *fs, myf MyFlags)
{
  int fd= fs->file_descr;
  os_off_t pos;
  DBUG_ENTER("my_simple_tell");
  DBUG_PRINT("my",("Fd: %d  MyFlags: %d",fd, MyFlags));
  
#ifdef HAVE_TELL
  pos= tell(fd);
#else
  pos= lseek(fd, 0L, MY_SEEK_CUR);
#endif
  if (pos == (os_off_t) -1)
    my_errno=errno;
  DBUG_PRINT("exit",("pos: %lu",pos));
  DBUG_RETURN((my_off_t) pos);
}

my_off_t my_simple_seek(MY_FILE * fs, my_off_t pos, int whence, myf MyFlags)
{
  int fd= fs->file_descr;
  reg1 os_off_t newpos;
  DBUG_ENTER("my_simple_seek");
  DBUG_PRINT("my",("Fd: %d  Hpos: %lu  Pos: %lu  Whence: %d  MyFlags: %d",
		   fd, ((ulonglong) pos) >> 32, (ulong) pos, whence, MyFlags));

  newpos= lseek(fd, pos, whence);
  if (newpos == (os_off_t) -1)
  {
    my_errno= errno;
    DBUG_PRINT("error",("lseek: %lu, errno: %d",newpos,errno));
    DBUG_RETURN(MY_FILEPOS_ERROR);
  }
  if ((my_off_t) newpos != pos)
  {
    DBUG_PRINT("exit",("pos: %lu", (ulong) newpos));
  }
  DBUG_RETURN((my_off_t) newpos);
}

int my_simple_delete(MY_FILE * fs, myf MyFlags)
{
  int err;
  char file_name[FN_REFLEN];
  char *file_for_delete;

  DBUG_ENTER("my_simple_delete");
  DBUG_PRINT("my",("name %s MyFlags %d", fs->name.buffer, MyFlags));

  file_for_delete= (char *)system_my_filename(file_name, fs);

  if ((err = unlink(file_for_delete)) == -1)
  {
    my_errno=errno;
    if (MyFlags & (MY_FAE+MY_WME))
      my_error(EE_DELETE,MYF(ME_BELL+ME_WAITTANG+(MyFlags & ME_NOINPUT)),
	       fs->name.buffer, errno);
  }
  DBUG_RETURN(err);
}
